=begin
 This file is part of Hopsworks
 Copyright (C) 2024, Hopsworks AB. All rights reserved

 Hopsworks is free software: you can redistribute it and/or modify it under the terms of
 the GNU Affero General Public License as published by the Free Software Foundation,
 either version 3 of the License, or (at your option) any later version.

 Hopsworks is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License along with this program.
 If not, see <https://www.gnu.org/licenses/>.
=end

describe "On #{ENV['OS']}" do
  before :all do
    @debugOpt=false
    # Enable the reporting
    RSpec.configure do |c|
      c.example_status_persistence_file_path  = 'execution_debug.txt' if @debugOpt
    end
  end
  after(:all) {clean_all_test_projects(spec: "ray_execution")}

  describe "with ray jobs" do
    before(:all) do
      if !ray_enabled
        skip "Ray jobs not enabled moving on...."
      end
      with_valid_project
    end

    describe 'execution core' do
      context 'create' do
        context 'without authentication' do
          before :all do
            with_valid_project
            reset_session
          end

          it "should fail" do
            get_job(@project[:id], 1, expected_status: 401, error_code: 200003)
          end
        end
        job_types = ['py', 'ipynb']
        job_types.each do |type|
          context 'with authentication and executable ' + type do
            before :each do
              reset_session
              create_session(@project[:username], "Pass123") #login as project owner
            end
            after :each do
              clean_jobs(@project[:id])
            end

            it "should start a job and get its executions" do
              #create job
              $job_name_1 = "demo_job_1_" + type
              create_ray_job(@project, $job_name_1, type:type)
              job_id = json_body[:id]
              #start execution
              start_execution(@project[:id], $job_name_1)
              execution_id = json_body[:id]
              begin
                expect(json_body[:state]).to eq "INITIALIZING"
                #get execution
                get_execution(@project[:id], $job_name_1, json_body[:id])
                expect(json_body[:id]).to eq(execution_id)
              ensure
                #wait till it's finished and start second execution
                wait_for_execution_completed(@project[:id], $job_name_1, json_body[:id], "FINISHED")
              end
              #start execution
              start_execution(@project[:id], $job_name_1)
              execution_id = json_body[:id]
              begin
                #get all executions of job
                get_executions(@project[:id], $job_name_1)
                expect(json_body[:items].count).to eq 2

                #check database
                num_executions = count_executions(job_id)
                expect(num_executions).to eq 2
              ensure
                wait_for_execution_completed(@project[:id], $job_name_1, execution_id, "FINISHED")
              end
            end
            it "should start and stop job" do
              $job_name_2 = "demo_job_2_" + type
              create_ray_job(@project, $job_name_2, type:type)

              #start execution
              start_execution(@project[:id], $job_name_2)
              execution_id = json_body[:id]
              wait_for_execution_active(@project[:id], $job_name_2, execution_id, "ACCEPTED", "appId")
              stop_execution(@project[:id], $job_name_2, execution_id)
            end
            it "should start two executions in parallel" do
              $job_name_3 = "demo_job_3_" + type
              create_ray_job(@project, $job_name_3, type:type)
              begin
                start_execution(@project[:id], $job_name_3)
                execution_id1 = json_body[:id]
                start_execution(@project[:id], $job_name_3)
                execution_id2 = json_body[:id]
              ensure
                wait_for_execution_completed(@project[:id], $job_name_3, execution_id1, "FINISHED") unless execution_id1.nil?
                wait_for_execution_completed(@project[:id], $job_name_3, execution_id2, "FINISHED") unless execution_id2.nil?
              end
            end
            it "should start a job, delete it and cleanup its executions" do
              #create job
              job_name = "demo_job"
              create_ray_job(@project, $job_name_1, type:type)
              job_id = json_body[:id]
              #start execution
              start_execution(@project[:id], job_id)
              execution_id = json_body[:id]
              #wait till it's finished and start second execution
              wait_for_execution_active(@project[:id], job_name, execution_id, "SUBMITTED", "appId")
              #start execution
              start_execution(@project[:id], job_name)
              execution_id = json_body[:id]
              wait_for_execution_active(@project[:id], job_name, execution_id, "SUBMITTED", "appId")

              # Delete job
              delete_job(@project[:id], job_name)
              #Wait for executions to be deleted
              sleep(getVar("executions_cleaner_interval_ms").value.to_i * 2/1000)

              #check database
              num_executions = count_executions(job_id)
              expect(num_executions).to eq 0
            end
            it "should start an execution and delete it while running" do
              job_name = "demo_job_5_" + type
              create_sparktour_job(@project, job_name, type)
              #start execution
              start_execution(@project[:id], job_name)
              execution_id = json_body[:id]
              wait_for_execution_active(@project[:id], job_name, execution_id, "SUBMITTED", "appId")
              get_execution(@project[:id], job_name, execution_id)
              app_id = json_body[:appId]
              delete_execution(@project[:id], job_name, execution_id)

              #check database
              num_executions = count_executions(job_name)
              expect(num_executions).to eq 0
            end
            def get_logs(project, job_name, execution_id, application_id, log_type, job_type)
              wait_for_me_time(60) do
                get_execution_log(project[:id], job_name, execution_id, log_type)
                found_state = json_body.key?(:log) && json_body[:log] != "No log available."
                { 'success' => found_state, 'msg' => "Timed-out waiting for #{log_type} log aggregation", "result" => json_body}
              end
              expect_status_details(200)

              unless check_log_dir_has_files(application_id, project[:projectname], job_type)
                raise "Log aggregation failed."
              end
              expect(json_body[:log]).not_to eq("No log available.")
              pp "execution json_body: #{json_body}" if defined?(@debugOpt) && @debugOpt
              expect(json_body[:type]).to eq log_type.upcase
            end
            def run_job_and_get_logs(project, job_name, type)
              start_execution(project[:id], job_name)
              execution_id = json_body[:id]
              hdfs_user = json_body[:hdfsUser]
              wait_for_execution_completed(project[:id], job_name, execution_id, "FINISHED")
              get_execution(project[:id], job_name, execution_id)
              application_id = json_body[:appId]
              job_type = "PySpark"
              if type == "jar"
                job_type = "Spark"
              end

              get_logs(project, job_name, execution_id, application_id, "out", job_type)
              get_logs(project, job_name, execution_id, application_id,  "err", job_type)
            end
            it "should run a job and get out and err logs" do
              $job_name_4 = "demo_job_4_" + type
              create_ray_job(@project, $job_name_4, type: type)
              create_ray_job(@project, $job_name_4, type: type)

              member = create_user
              add_member_to_project(@project, member[:email], "Data scientist")
              create_session(member[:email], "Pass123")
              run_job_and_get_logs(@project, $job_name_4, type)
            end
          end
        end
      end
    end
    describe 'execution extended' do
      context 'create' do
        job_types = ['py', 'ipynb']
        job_types.each do |type|
          context 'with authentication and executable ' + type do
            before :each do
              $job_name = "j_#{short_random_id}"
            end
            after :each do
              clean_jobs(@project[:id])
            end
            it "should start a job and use default args" do
              create_ray_job(@project, $job_name, type: type)
              default_args = json_body[:config][:defaultArgs]
              start_execution(@project[:id], $job_name)
              execution_id = json_body[:id]
              begin
                get_execution(@project[:id], $job_name, execution_id)
                expect(json_body[:args]).not_to be_nil
                expect(default_args).not_to be_nil
                expect(json_body[:args]).to eq default_args
              ensure
                wait_for_execution_completed(@project[:id], $job_name, execution_id, "FINISHED")
              end
            end
            it "should start a ray job with args 123" do
              create_ray_job(@project, $job_name, type: type)
              args = "123"
              start_execution(@project[:id], $job_name, args: args)
              execution_id = json_body[:id]
              begin
                get_execution(@project[:id], $job_name, execution_id)
                expect(json_body[:args]).to eq args
              ensure
                wait_for_execution_completed(@project[:id], $job_name, execution_id, "FINISHED")
              end
            end
            it "should fail to start a ray job with missing appPath" do
              create_ray_job(@project, $job_name, type: type)
              config = json_body[:config]

              create_ray_job(@project, $job_name, type: type, job_conf: config, expected_status: 200)
              delete_dataset(@project, config[:appPath], datasetType: "?type=DATASET")
              #start execution
              start_execution(@project[:id], $job_name, expected_status: 400)
            end
            it "should fail to start a ray job with missing runtime environment" do
              create_ray_job(@project, $job_name, type: type)
              config = json_body[:config]
              delete_dataset(@project, config[:runtimeEnvironment], datasetType: "?type=DATASET")
              #start execution
              start_execution(@project[:id], $job_name, expected_status: 400)
            end
            it 'should fail to start a ray job with missing files' do
              create_ray_job(@project, $job_name, type: type)
              config = json_body[:config]
              config[:files] = "/Projects/#{@project[:projectname]}/Resources/test_ray_file.py"
              create_ray_job(@project, $job_name, type: type, job_conf: config, expected_status: 200)
              delete_dataset(@project, config[:files], datasetType: "?type=DATASET")
              #start execution
              start_execution(@project[:id], $job_name, expected_status: 400)
            end
          end
        end
      end

      describe "#access" do
        def run_job(user, project, job_name)
          begin
            get_job(project["id"], job_name, expected_status: 200)
          rescue RSpec::Expectations::ExpectationNotMetError => e
            if job_does_not_exist
              create_ray_job(project, job_name)
              start_execution(project["id"], job_name)
            end
          end
        end
        def get_execution_(project, job_name)
          get_executions(project["id"], job_name)
          expect(json_body[:items].length).to be > 0
          json_body[:items][0]
        end

        describe '#access' do
          before :all do
            @user_data_owner = create_user
            pp @user_data_owner[:email] if defined?(@debugOpt) && @debugOpt
            @user_data_scientist = create_user
            @user_other = create_user
            create_session(@user_data_owner[:email], "Pass123")
            @project = create_project
            add_member_to_project(@project, @user_data_scientist[:email], "Data scientist")
            @job_name = "test_job_#{short_random_id}"
            run_job(@user_data_owner, @project, @job_name)
            get_executions(@project["id"], @job_name)
            execution_id = json_body[:items][0][:id]
            app_id = json_body[:items][0][:appId]
            wait_for_execution_active(@project[:id], @job_name, execution_id, "FINISHED", "appId")
          end

          it 'data owner should see ray dashboard'  do
            raw_create_session(@user_data_owner[:email], "Pass123")
            execution = get_execution_(@project, @job_name)
            get "/hopsworks-api/#{execution[:monitoring][:rayUrl]}"
            expect_status_details(200)
          end
          it 'data scientist should see ray dashboard' do
            raw_create_session(@user_data_scientist[:email], "Pass123")
            execution = get_execution_(@project, @job_name)
            get "/hopsworks-api/#{execution[:monitoring][:rayUrl]}"
            expect_status_details(200)
          end
          it 'other user should not see ray dashboard' do
            raw_create_session(@user_data_owner[:email], "Pass123")
            execution = get_execution_(@project, @job_name)
            raw_create_session(@user_other[:email], "Pass123")
            get "/hopsworks-api/#{execution[:monitoring][:rayUrl]}"
            expect_status_details(400)
          end
        end
      end

    end
  end
end