=begin
 This file is part of Hopsworks
 Copyright (C) 2024, Hopsworks AB. All rights reserved

 Hopsworks is free software: you can redistribute it and/or modify it under the terms of
 the GNU Affero General Public License as published by the Free Software Foundation,
 either version 3 of the License, or (at your option) any later version.

 Hopsworks is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License along with this program.
 If not, see <https://www.gnu.org/licenses/>.
=end
describe "On #{ENV['OS']}" do
  after(:all) {clean_all_test_projects(spec: "ray_job")}
  job_ray_1 = "demo_ray_job_1"
  job_ray_2 = "demo_ray_job_2"
  describe 'job' do
    context 'without authentication' do
      before :all do
        if !ray_enabled
          skip "Ray jobs not enabled moving on...."
        end
        with_valid_project
        reset_session
      end
      it "should fail" do
        get_job(@project[:id], 1, expected_status: 401, error_code: 200003)
      end
    end
    context 'with authentication create, delete, get' do
      before :all do
        with_valid_project
      end
      it "should create a ray job" do
        create_ray_job(@project, job_ray_1)
      end

      describe 'create, update ray job' do
        job_name = job_ray_2
        before :all do
          create_ray_job(@project, job_name)
          get_job(@project[:id], job_name)
          expect_status_details(200)
        end
        it "should update a job with different job_config" do
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'workerMaxInstances'] = '2'
          create_ray_job(@project, job_name, job_conf: config, expected_status: 200)
          get_job(@project[:id], job_name)
          expect_status_details(200)
          expect(json_body[:config][:'workerMaxInstances']).to eq 2
        end

        it "should fail to create or update job if worker min instances is less than 0" do
          # get job, change args and config params and put it
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'workerMinInstances'] = '-1'
          create_ray_job(@project, job_name, job_conf: config, expected_status: 422)
        end

        it "should fail to create or update job if worker min instances is greater than worker max instances" do
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'workerMinInstances'] = '10'
          create_ray_job(@project, job_name, job_conf: config, expected_status: 422)
        end

        it "should fail to create or update job if worker max instances is 0" do
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'workerMaxInstances'] = '0'
          create_ray_job(@project, job_name, job_conf: config, expected_status: 422)
        end

        it "should fail to create or update job if worker max instances is greater than the possible worker max instances" do
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'workerMaxInstances'] = '40'
          create_ray_job(@project, job_name, job_conf: config, expected_status: 422)
        end

        it "should fail to create or update job if main script is not set" do
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'appPath'] = ''
          json_result = create_ray_job(@project, job_name, job_conf: config, expected_status: 422)
          parsed_json = JSON.parse(json_result.body)
          expect(parsed_json["usrMsg"] == "Path does not point to a .py or .ipynb file.").to be true
        end

        it 'should fail to create or update job app file does not exist' do
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'appPath'] = "hdfs:///Projects/#{@project[:projectname]}/Resources/not_existent_ray_file.py"
          json_result = create_ray_job(@project, job_name, job_conf: config, expected_status: 400)
          parsed_json = JSON.parse(json_result.body)
          expect(parsed_json["errorCode"] == 130047).to be true
          expect((parsed_json["usrMsg"]).include?("Job application file does not exist")).to be true
        end

        it 'should fail to create or update job if attached dependancy does not exist' do
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'files'] = "hdfs:///Projects/#{@project[:projectname]}/Resources/not_existent_file.py"
          json_result = create_ray_job(@project, job_name, job_conf: config, expected_status: 400)
          parsed_json = JSON.parse(json_result.body)
          expect(parsed_json["errorCode"] == 130047).to be true
          expect((parsed_json["usrMsg"]).include?("Attached file does not exist")).to be true
        end

        it "should fail to create or update job if runtime environment file is invalid yaml" do
          runtime_env_filename = "invalid_ray_runtime_env.yaml"
          copy_from_local("#{ENV['PROJECT_DIR']}/hopsworks-IT/src/test/ruby/spec/auxiliary/#{runtime_env_filename}",
                          "/Projects/#{@project[:projectname]}/Resources", @user[:username], "#{@project[:projectname]}__Resources", 750, "#{@project[:projectname]}")

          get "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/dataset/Resources/?action=listing&expand=inodes"
          expect_status_details(200)
          yaml_file = json_body[:items].detect { |d| d[:attributes][:name] == "#{runtime_env_filename}" }
          expect(yaml_file).to be_present
          get_job(@project[:id], job_name)
          expect_status_details(200)
          config = json_body[:config]
          config[:'runtimeEnvironment'] = "hdfs:///Projects/#{@project[:projectname]}/Resources/#{runtime_env_filename}"
          json_result = create_ray_job(@project, job_name, job_conf: config, expected_status: 400)
          parsed_json = JSON.parse(json_result.body)
          expect(parsed_json["errorCode"] == 130045).to be true
          expect((parsed_json["usrMsg"]).include?("The environment yaml file provided does not have valid yaml")).to be true
        end
      end
    end
  end
end